"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const dataloader_1 = __importDefault(require("dataloader"));
const redis_commands_1 = require("redis-commands");
const scan_stream_1 = require("./scan-stream");
const stats_1 = require("./stats");
exports.BatchStats = stats_1.BatchStats;
const scan_iterable_1 = require("./scan-iterable");
class RedisLoader {
    constructor({ redis, logger, maxBatchSize = Infinity }) {
        if (!redis) {
            throw new Error('"redis" is required');
        }
        this.logger = logger;
        this.stats = new stats_1.RedisStats();
        this.redis = redis;
        this.dataLoader = new dataloader_1.default(commands => this.batchFunction(commands), { cache: false, maxBatchSize });
    }
    resetStats() {
        this.stats = new stats_1.RedisStats();
    }
    get _redis() {
        return this.redis;
    }
    startBatch(batchInfo) {
        return this.stats.startBatch(batchInfo);
    }
    endBatch(batchStats, error, response) {
        this.stats.endBatch(batchStats, error, response);
        this.logStats();
    }
    logStats() {
        if (this.logger) {
            this.logger(this.stats);
        }
    }
    batchFunction(commands) {
        if (!(commands && commands.length > 0)) {
            throw new Error('no commands to run');
        }
        if (commands.length === 1) {
            return this.executeSingle(commands);
        }
        return this.executeMulti(commands);
    }
    async executeMulti(commands) {
        const batchStats = this.startBatch({ commands, multi: true });
        try {
            const response = await this.redis.multi(commands).exec();
            this.endBatch(batchStats, null, response);
            return response.map(([error, data]) => error || data);
        }
        catch (error) {
            if (Array.isArray(error.previousErrors)) {
                error.message = `${error.message}: ${error.previousErrors.map(e => e && e.message)}`;
            }
            this.endBatch(batchStats, error, null);
            throw error;
        }
    }
    async executeSingle(commands) {
        const batchStats = this.startBatch({ commands, multi: false });
        const [command, ...args] = commands[0];
        try {
            const response = await this.redis[command](...args);
            this.endBatch(batchStats, null, [response]);
            return [response];
        }
        catch (error) {
            this.endBatch(batchStats, error, null);
            throw error;
        }
    }
}
exports.RedisLoader = RedisLoader;
// Dynamically create the redis functions on our prototype
// https://github.com/luin/ioredis/blob/58dc63006fcc98f7a13f363f920856afa887801e/lib/commander.js#L28
const redisCommands = new Set(redis_commands_1.list);
redisCommands.delete('monitor');
redisCommands.add('sentinel');
redis_commands_1.list.forEach(command => {
    RedisLoader.prototype[command] = function (...args) {
        return this.dataLoader.load([command, ...args]);
    };
    const bufferCmd = `${command}Buffer`;
    RedisLoader.prototype[bufferCmd] = function (...args) {
        return this.dataLoader.load([bufferCmd, ...args]);
    };
});
const scanCommands = ['scan', 'sscan', 'hscan', 'zscan', 'scanBuffer', 'sscanBuffer', 'hscanBuffer', 'zscanBuffer'];
scanCommands.forEach(command => {
    RedisLoader.prototype[`${command}Stream`] = function (key, options) {
        if (command === 'scan' || command === 'scanBuffer') {
            options = key;
            key = null;
        }
        return new scan_stream_1.ScanStream(Object.assign({ key, redis: this, command }, options));
    };
    RedisLoader.prototype[`${command}Iterable`] = function (key, options) {
        if (command === 'scan' || command === 'scanBuffer') {
            options = key;
            key = null;
        }
        return scan_iterable_1.scanIterable(Object.assign({ key, redis: this, command }, options));
    };
});
const pubSubCommands = ['subscribe', 'unsubscribe', 'publish', 'psubscribe', 'punsubscribe'];
pubSubCommands.forEach(command => {
    RedisLoader.prototype[command] = function (...args) { return this.redis[command](...args); };
    RedisLoader.prototype[`${command}Buffer`] = function (...args) { return this.redis[`${command}Buffer`](...args); };
});
const passThroughCommands = ['connect', 'disconnect'];
passThroughCommands.forEach(command => {
    RedisLoader.prototype[command] = function (...args) { return this.redis[command](...args); };
});
Object.keys(events_1.EventEmitter.prototype).forEach(key => {
    if (typeof events_1.EventEmitter.prototype[key] === 'function') {
        RedisLoader.prototype[key] = function (...args) {
            return this.redis[key](...args);
        };
    }
    else {
        Object.defineProperty(RedisLoader.prototype, key, {
            get() {
                return this.redis[key];
            }
        });
    }
});
